"""Nile plugin adding coverage reports for Cairo Smart Contracts."""
import logging

try:
    from importlib import metadata as importlib_metadata
except ImportError:
    import importlib_metadata

try:
    __version__ = importlib_metadata.version("nile-coverage")
except importlib_metadata.PackageNotFoundError:
    __version__ = None


__author__ = "Eric Nordelo"
__license__ = "MIT"
__url__ = "https://github.com/ericnordelo/nile-coverage"

logger = logging.getLogger(__name__)



"""Tests for coverage command."""

import logging

from click.testing import CliRunner

from nile_coverage import __name__
from nile_coverage.commands import coverage


def test_coverage(caplog):
    """Command tests."""
    caplog.set_level(logging.INFO, logger=__name__)

    CliRunner().invoke(coverage)

    assert "Generating coverage report" in caplog.text
    assert "Nothing to report" in caplog.text



"""Coverage command definition."""

import click
import pytest

from nile_coverage import logger
from nile_coverage.plugins import CoveragePlugin


@click.command()
@click.option("--mark", "-m", help="Pytest mark wrapper.")
@click.option(
    "--single-thread", "-s", is_flag=True, help="Run the test suite in a single thread."
)
@click.option(
    "--contracts-folder",
    "-c",
    default="contracts",
    help="Specify contracts folder (default to 'contracts').",
)
@click.option(
    "--xml", is_flag=True, help="Create a coverage.xml report with Cobertura format."
)
def coverage(mark, single_thread, contracts_folder, xml):
    """Generate coverage report for Cairo Smart Contracts."""
    logger.info("\nGenerating coverage report")

    args = ["-p", "no:warnings"]

    if mark is not None:
        args += ["-m", mark]

    if single_thread:
        args += ["-n", "0"]

    pytest.main(args, plugins=[CoveragePlugin(contracts_folder, xml)])



from collections import defaultdict
from dataclasses import dataclass
import os

from textwrap import wrap
from typing import Any, DefaultDict, Dict, List, Optional, Set

from starkware.cairo.lang.compiler.instruction import Instruction
from starkware.cairo.lang.compiler.program import ProgramBase
from starkware.cairo.lang.vm import cairo_runner
from starkware.cairo.lang.vm.builtin_runner import BuiltinRunner
from starkware.cairo.lang.vm.relocatable import MaybeRelocatable
from starkware.cairo.lang.vm.vm_core import RunContext, VirtualMachine

from nile_coverage import logger
from nile_coverage.vendor.reporters import XmlReporter


class Headers:
    """Headers for the report table."""

    FILE: str = "File "
    FILE_INDEX: int = 0

    COVERED: str = "Covered(%) "
    COVERED_INDEX: int = 1

    MISSED: str = "Missed(%) "
    MISSED_INDEX: int = 2

    LINES_MISSED: str = "Lines missed"
    LINE_MISSED_INDEX: int = 3


class Colors:
    """Colors to indicate if the coverage is good or not."""

    FAIL = "\033[91m"
    GREEN = "\033[92m"
    WARNING = "\033[93m"
    END = "\033[0m"


@dataclass
class CoverageFile:
    name: str  # Filename.
    covered: Set[int]  # Tested lines.
    statements: Set[int]  # Lines with code.
    precision: int = 1  # Decimals for %.

    @staticmethod
    def col_sizes(sizes=[]):
        """To share the column sizes between all the instances."""
        return sizes

    def __post_init__(self):
        self.nb_statements = len(
            self.statements
        )  # Nb of lines with code in the cairo file.
        self.nb_covered = len(self.covered)  # Nb of lines tested.
        self.missed = sorted(list(self.statements - self.covered))  # Lines not tested.
        self.nb_missed = len(self.missed)  # Nb of lines not tested.
        self.pct_covered = 0 if self.nb_statements == 0 else (
            100 * self.nb_covered / self.nb_statements
        )  # % of lines tested.
        self.pct_missed = 100 if self.nb_statements == 0 else (
            100 * self.nb_missed / self.nb_statements
        )  # % of lines not tested.

    def __str__(self):
        sizes = self.__class__.col_sizes()  # Get columns size.
        name_len = len(self.name)
        if (
            name_len > sizes[Headers.FILE_INDEX]
        ):  # If the filename is longer than the col len we crop it.
            cropped_name = self.name[
                5 + name_len - sizes[Headers.FILE_INDEX] :
            ]  # Crops the filename to the filename col size and leave room for [...] prefix.
            # Formats the string to be left aligned and to file the col size length.
            name = f"{'[...]' + cropped_name:<{sizes[Headers.FILE_INDEX]}}"
        else:
            # Pads the filename to the column length.
            name = f"{self.name:<{sizes[Headers.FILE_INDEX]}}"
        # % covered centered with right decimals.
        pct_covered = (
            f"{self.pct_covered:^{sizes[Headers.COVERED_INDEX]}.{self.precision}f}"
        )
        # % missed centered with right decimals.
        pct_missed = (
            f"{self.pct_missed:^{sizes[Headers.MISSED_INDEX]}.{self.precision}f}"
        )
        prefix = " " * (
            len(name) + len(pct_covered) + len(pct_missed) + 4
        )  # Offset of the missed lines column.
        if self.pct_covered == 0:
            missed = "[...]"
        elif len(str(self.missed)) > sizes[Headers.LINE_MISSED_INDEX]:
            wrapped_missed = wrap(
                str(self.missed), sizes[Headers.LINE_MISSED_INDEX] - 1
            )  # Wrap the missed lines list if too big.
            wrapped_missed[1:] = [
                f"{prefix}{val}" for val in wrapped_missed[1:]
            ]  # Prefix the wrapped missed lines.
            missed: str = "\n".join(wrapped_missed)  # Convert it to multiline string.
        else:
            missed = str(self.missed)
        if 0 <= self.pct_covered < 50:  # If coverage is not enough writes in red.
            color = Colors.FAIL
        elif 50 <= self.pct_covered < 80:  # If coverage is mid enough writes in yellow.
            color = Colors.WARNING
        else:  # If coverage is good write in green.
            color = Colors.GREEN
        # Formatted file line report.
        return f"{color}{name} {pct_covered} {pct_missed} {missed}{Colors.END}"


def print_sum(covered_files: List[CoverageFile]):
    """Print the coverage summary of the project."""
    try:
        term_size = os.get_terminal_size()
        max_name = max([len(file.name) for file in covered_files]) + 2  # Longest name.
        max_missed_lines = max(
            [len(str(file.missed)) for file in covered_files]
        )  # Length of the longest missed lines list.
        sizes = (
            CoverageFile.col_sizes()
        )  # Init the sizes list with our static method so it's available everywhere.
        sizes.extend(
            [max_name, len(Headers.COVERED), len(Headers.MISSED), max_missed_lines]
        )  # Fill the sizes.

        while (
            sum(sizes) > term_size.columns
        ):  # While the length of all the cols is > the terminal size, reduce the biggest col.
            idx = sizes.index(max(sizes))
            sizes[idx] = int(0.75 * sizes[idx])

        headers = (
            f"\n\n{Headers.FILE:{sizes[Headers.FILE_INDEX] + 1}}"
            f"{Headers.COVERED:{sizes[Headers.COVERED_INDEX] + 1}}"
            f"{Headers.MISSED:{sizes[Headers.MISSED_INDEX] + 1}}"
            f"{Headers.LINES_MISSED:{sizes[Headers.LINE_MISSED_INDEX] + 1}}\n"
        )  # Prepare the coverage table headers.

        total_covered = 0
        underline = "-" * term_size.columns  # To separate the header from the values.
        logger.info(headers + underline)
        for file in covered_files:  # Prints the report of each file.
            logger.info(file)
            total_covered += file.pct_covered

        total_str = "TOTAL"
        total_covered /= len(covered_files)
        footer = (
            f"\n{total_str:{sizes[Headers.FILE_INDEX] + 1}}"
            f"{total_covered:{8}}\n"
        )  # Prepare the coverage table footer.

        logger.info(underline + footer)
    except OSError:
        pass


def report_runs(
    contracts_folder: str = "",
    print_summary: bool = True,
):
    if not os.path.isdir(contracts_folder):
        logger.info(f"\n\nNothing to report (couldn't find \"{contracts_folder}\" directory)")
        return "Nothing to report"

    report_dict = OverrideVm.covered()  # Get the infos of all the covered files.
    statements = OverrideVm.statements()  # Get the lines of codes of each files.

    # Add all the files to the report (reporting 0 coverage files)
    add_files_to_report(contracts_folder, report_dict)

    files = sorted(
        [
            CoverageFile(
                statements=set(statements[file]), covered=set(coverage), name=file
            )
            for file, coverage in report_dict.items()
            if file.startswith(contracts_folder)
        ],
        key=lambda x: x.name,
    )  # Sort the files by filename.

    if not len(files):
        logger.info("\n\nNothing to report")
        return "Nothing to report"
    if print_summary:
        print_sum(covered_files=files)
    return files

def report_runs_xml(
    contracts_folder: str = ""
):
    if not os.path.isdir(contracts_folder):
        logger.info(f"\n\nNothing to report (couldn't find \"{contracts_folder}\" directory)")
        return "Nothing to report"

    report_dict = OverrideVm.covered()  # Get the infos of all the covered files.
    statements = OverrideVm.statements()  # Get the lines of codes of each files.

    reporter = XmlReporter(contracts_folder, statements, report_dict)
    reporter.report(outfile="coverage.xml")


def add_files_to_report(contracts_folder: str, report_dict: DefaultDict[str, List[int]]):
    for path, subdirs, files in os.walk(contracts_folder):
        for name in files:
            f = os.path.join(path, name)
            if f not in report_dict and f.endswith('.cairo'):
                report_dict[f] = []


def process_file(file: str):
    cwd = os.getcwd()
    if file.startswith(cwd):
        return file.removeprefix(cwd+'/')
    return file


class OverrideVm(VirtualMachine):
    def __init__(
        self,
        program: ProgramBase,
        run_context: RunContext,
        hint_locals: Dict[str, Any],
        static_locals: Optional[Dict[str, Any]] = None,
        builtin_runners: Optional[Dict[str, BuiltinRunner]] = None,
        program_base: Optional[MaybeRelocatable] = None,
    ):
        super().__init__(
            program=program,
            run_context=run_context,
            hint_locals=hint_locals,
            static_locals=static_locals,
            builtin_runners=builtin_runners,
            program_base=program_base,
        )
        self.old_end_run = (
            super().end_run
        )  # Save the old end run function to wrap it afterwards.
        self.old_run_instruction = (
            super().run_instruction
        )  # Save the old run instruction function to wrap it afterwards.
        self.old_as_vm_exception = (
            super().as_vm_exception
        )  # Save the old vm as exception function to wrap it afterwards.
        self.touched_pcs: List[int] = []

    def run_instruction(self, instruction: Instruction):
        """Saves the current pc and runs the instruction."""
        self.touched_pcs.append(self.run_context.pc.offset)
        self.old_run_instruction(instruction=instruction)

    def end_run(self):
        """In case the run doesn't fail creates report coverage."""
        self.old_end_run()
        self.cover_file()

    def as_vm_exception(
        self,
        exc,
        with_traceback: bool = True,
        notes: Optional[List[str]] = None,
        hint_index: Optional[int] = None,
    ):
        """In case the run fails creates report coverage."""
        self.cover_file()
        return self.old_as_vm_exception(exc, with_traceback, notes, hint_index)

    @staticmethod
    def covered(
        val: DefaultDict[str, List[int]] = defaultdict(list)
    ) -> DefaultDict[str, list]:
        """To share the covered files between all the instances."""
        return val

    @staticmethod
    def statements(
        val: DefaultDict[str, List[int]] = defaultdict(list)
    ) -> DefaultDict[str, list]:
        """To share the lines of codes in files between all the instances."""
        return val

    def pc_to_line(
        self,
        pc,
        statements: DefaultDict[str, list],
        report_dict: DefaultDict[str, List[int]],
    ) -> None:
        """Converts the touched pcs to the line numbers of the original file and saves them."""
        should_update_report = (
            pc in self.touched_pcs
        )  # If the pc is not touched by the test don't report it.
        instruct = self.program.debug_info.instruction_locations[
            pc
        ].inst  # First instruction in the debug info.
        file = process_file(instruct.input_file.filename)  # Current analyzed file relative path.
        while True:
            if "autogen" not in file:  # If file is auto generated discard it.
                lines = list(
                    range(
                        instruct.start_line,
                        instruct.end_line + 1,
                    )
                )  # Get the lines touched.
                if should_update_report:
                    report_dict[file].extend(lines)
                statements[file].extend(lines)
            if (
                instruct.parent_location is not None
            ):  # Continue until we have last parent location.
                instruct = instruct.parent_location[0]
                file = instruct.input_file.filename
            else:
                return

    def cover_file(
        self,
    ):
        """Adds the coverage report in the report dict and all the lines of code."""
        if self.program.debug_info is not None:
            report_dict = self.__class__.covered()
            statements = self.__class__.statements()
            for pc in set(self.program.debug_info.instruction_locations.keys()):
                self.pc_to_line(pc=pc, report_dict=report_dict, statements=statements)


cairo_runner.VirtualMachine = OverrideVm



"""Coverage reporters"""

import os
import os.path
import sys
import time
import xml.dom.minidom

from nile_coverage import __version__, __url__
from nile_coverage.vendor.coverage_file import CoverageFile, add_files_to_report


def rate(hit, num):
    """Return the fraction of `hit`/`num`, as a string."""
    if num == 0:
        return "1"
    else:
        return "%.4g" % (float(hit) / num)


class XmlReporter:
    """Cobertura-style XML reports."""

    def __init__(self, contracts_folder, statements, report_dict):
        self.statements = statements
        self.report_dict = report_dict
        self.contracts_folder = contracts_folder

        # add empty coverage files to report
        add_files_to_report(contracts_folder, self.report_dict)

        self.source_paths = set()
        self.packages = {}
        self.xml_out = None

    def report(self, outfile=None):
        """Generate a Cobertura-compatible XML report for `morfs`.
        `morfs` is a list of modules or file names.
        `outfile` is a file object to write the XML to.
        """
        # Initial setup.
        outfile = outfile or sys.stdout

        # Create the DOM that will store the data.
        impl = xml.dom.minidom.getDOMImplementation()
        self.xml_out = impl.createDocument(None, "coverage", None)

        # Write header stuff.
        xcoverage = self.xml_out.documentElement
        xcoverage.setAttribute("version", __version__)
        xcoverage.setAttribute("timestamp", str(int(time.time()*1000)))
        xcoverage.appendChild(self.xml_out.createComment(
            f" Generated by nile-coverage: {__url__} "
        ))

        # Call xml_file for each file in the data.
        for file, coverage in self.report_dict.items():
            if file.startswith(self.contracts_folder):
                self.xml_file(CoverageFile(
                    statements=set(self.statements[file]), covered=set(coverage), name=file
                ))

        xsources = self.xml_out.createElement("sources")
        xcoverage.appendChild(xsources)

        # Populate the XML DOM with the source info.
        for path in sorted(self.source_paths):
            xsource = self.xml_out.createElement("source")
            xsources.appendChild(xsource)
            txt = self.xml_out.createTextNode(path)
            xsource.appendChild(txt)

        lnum_tot, lhits_tot = 0, 0
        bnum_tot, bhits_tot = 0, 0

        xpackages = self.xml_out.createElement("packages")
        xcoverage.appendChild(xpackages)

        # Populate the XML DOM with the package info.
        for pkg_name, pkg_data in sorted(self.packages.items()):
            class_elts, lhits, lnum, bhits, bnum = pkg_data
            xpackage = self.xml_out.createElement("package")
            xpackages.appendChild(xpackage)
            xclasses = self.xml_out.createElement("classes")
            xpackage.appendChild(xclasses)
            for _, class_elt in sorted(class_elts.items()):
                xclasses.appendChild(class_elt)
            xpackage.setAttribute("name", pkg_name.replace(os.sep, '.'))
            xpackage.setAttribute("line-rate", "0")
            branch_rate = "0"
            xpackage.setAttribute("branch-rate", branch_rate)
            xpackage.setAttribute("complexity", "0")

            lnum_tot += lnum
            lhits_tot += lhits
            bnum_tot += bnum
            bhits_tot += bhits

        xcoverage.setAttribute("lines-valid", str(lnum_tot))
        xcoverage.setAttribute("lines-covered", str(lhits_tot))
        xcoverage.setAttribute("line-rate", rate(lhits_tot, lnum_tot))
        xcoverage.setAttribute("branches-covered", "0")
        xcoverage.setAttribute("branches-valid", "0")
        xcoverage.setAttribute("branch-rate", "0")
        xcoverage.setAttribute("complexity", "0")

        # Write the output file.
        cwd = os.getcwd()
        output = os.path.join(cwd, outfile)
        print(output)
        with open(output, "w") as f:
            f.write(serialize_xml(self.xml_out))

        # Return the total percentage.
        denom = lnum_tot + bnum_tot
        if denom == 0:
            pct = 0.0
        else:
            pct = 100.0 * (lhits_tot + bhits_tot) / denom
        return pct

    def xml_file(self, cf: CoverageFile):
        """Add to the XML report for a single file."""

        # Create the 'lines' and 'package' XML elements, which
        # are populated later.  Note that a package == a directory.
        cf.name = cf.name.replace("\\", "/")
        self.source_paths.add(cf.name)

        rel_name = cf.name
        dirname = os.path.dirname(rel_name) or "."
        dirname = "/".join(dirname.split("/")[:5])
        package_name = dirname.replace("/", ".")

        package = self.packages.setdefault(package_name, [{}, 0, 0, 0, 0])

        xclass = self.xml_out.createElement("class")

        xclass.appendChild(self.xml_out.createElement("methods"))

        xlines = self.xml_out.createElement("lines")
        xclass.appendChild(xlines)

        xclass.setAttribute("name", os.path.relpath(rel_name, dirname))
        xclass.setAttribute("filename", rel_name.replace("\\", "/"))
        xclass.setAttribute("complexity", "0")

        # For each statement, create an XML 'line' element.
        for line in sorted(cf.statements):
            xline = self.xml_out.createElement("line")
            xline.setAttribute("number", str(line))

            # Q: can we get info about the number of times a statement is
            # executed?  If so, that should be recorded here.
            xline.setAttribute("hits", "1")
            xlines.appendChild(xline)

        class_branches = 0.0
        class_br_hits = 0.0
        branch_rate = "0"
        line_rate = (
            f"{cf.pct_covered:.{cf.precision}f}"
        )

        # Finalize the statistics that are collected in the XML DOM.
        xclass.setAttribute("line-rate", line_rate)
        xclass.setAttribute("branch-rate", branch_rate)

        package[0][rel_name] = xclass
        package[1] += cf.nb_covered
        package[2] += cf.nb_statements
        package[3] += class_br_hits
        package[4] += class_branches


def serialize_xml(dom):
    """Serialize a minidom node to XML."""
    return dom.toprettyxml()


from dataclasses import dataclass
from typing import Dict, List, Set
import os

@dataclass
class CoverageFile:
    name: str  # Filename.
    covered: Set[int]  # Tested lines.
    statements: Set[int]  # Lines with code.
    precision: int = 1  # Decimals for %.

    @staticmethod
    def col_sizes(sizes=[]):
        """To share the column sizes between all the instances."""
        return sizes

    def __post_init__(self):
        self.nb_statements = len(
            self.statements
        )  # Nb of lines with code in the cairo file.
        self.nb_covered = len(self.covered)  # Nb of lines tested.
        self.missed = sorted(list(self.statements - self.covered))  # Lines not tested.
        self.nb_missed = len(self.missed)  # Nb of lines not tested.
        self.pct_covered = 0 if self.nb_statements == 0 else (
            100 * self.nb_covered / self.nb_statements
        )  # % of lines tested.
        self.pct_missed = 100 if self.nb_statements == 0 else (
            100 * self.nb_missed / self.nb_statements
        )  # % of lines not tested.


def add_files_to_report(contracts_folder: str, report_dict):
    for path, subdirs, files in os.walk(contracts_folder):
        for name in files:
            f = os.path.join(path, name)
            if f not in report_dict and f.endswith('.cairo'):
                report_dict[f] = []



"""PyTest plugin to print coverage report."""
from nile_coverage.vendor import cairo_coverage


class CoveragePlugin:
    def __init__(self, contracts_folder, xml=False):
        self.contracts_folder = contracts_folder
        self.xml = xml

    def pytest_sessionfinish(self):
        """
        Called after whole test run finished, right before
        returning the exit status to the system.
        """
        if self.xml:
            cairo_coverage.report_runs_xml(self.contracts_folder)
        else:
            cairo_coverage.report_runs(self.contracts_folder)

